/*********************************************************************
MLDemos: A User-Friendly visualization toolkit for machine learning
Copyright (C) 2010  Basilio Noris
Contact: mldemos@b4silio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License,
version 3 as published by the Free Software Foundation.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/
#include "interfaceSVMClassifier.h"
#include <QPixmap>
#include <QBitmap>
#include <QPainter>

using namespace std;

ClassSVM::ClassSVM()
{
	params = new Ui::Parameters();
	params->setupUi(widget = new QWidget());
	connect(params->svmTypeCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(ChangeOptions()));
}

void ClassSVM::ChangeOptions()
{
	int C = params->svmCSpin->value();
	params->maxSVSpin->setEnabled(false);
	params->svmCSpin->setRange(0.0001, 1.0);
	params->svmCSpin->setSingleStep(0.0001);
	params->svmCSpin->setDecimals(4);
	if(C > 1) params->svmCSpin->setValue(0.001);
	switch(params->svmTypeCombo->currentIndex())
	{
	case 0: // C-SVM
		params->svmCSpin->setRange(0.1, 9999);
		params->svmCSpin->setSingleStep(1);
		params->svmCSpin->setDecimals(1);
		if(params->svmCSpin->value() < 1) params->svmCSpin->setValue(100);
		params->svmTypeLabel->setText("C");
		break;
	case 1: // Nu-SVM
		params->svmTypeLabel->setText("Nu");
		break;
	case 2: // RVM
		params->svmTypeLabel->setText("eps");
		break;
	case 3: // Pegasos
		params->svmTypeLabel->setText("lambda");
		params->maxSVSpin->setEnabled(true);
		break;
	}
}

void ClassSVM::SetParams(Classifier *classifier)
{
	if(!classifier) return;
	int svmType = params->svmTypeCombo->currentIndex();
	int maxSV = params->maxSVSpin->value();
	float svmC = params->svmCSpin->value();
	int kernelType = params->kernelTypeCombo->currentIndex();
	float kernelGamma = params->kernelWidthSpin->value();
	float kernelDegree = params->kernelDegSpin->value();

	switch(classifier->type)
	{
	case CLASS_RVM:
		{
			ClassifierRVM *rvm = ((ClassifierRVM *)classifier);
			rvm->SetParams(svmC, kernelType, kernelGamma, kernelDegree);
		}
		break;
	case CLASS_PEG:
		{
			ClassifierPegasos *pegasos = ((ClassifierPegasos *)classifier);
			pegasos->SetParams(svmC, max(2,(int)maxSV), kernelType, kernelGamma, kernelDegree);
		}
	case CLASS_SVM:
		{
			ClassifierSVM *svm = ((ClassifierSVM *)classifier);
			switch(svmType)
			{
			case 0:
				svm->param.svm_type = C_SVC;
				break;
			case 1:
				svm->param.svm_type = NU_SVC;
				break;
			}
			switch(kernelType)
			{
			case 0:
				svm->param.kernel_type = LINEAR;
				break;
			case 1:
				svm->param.kernel_type = POLY;
				break;
			case 2:
				svm->param.kernel_type = RBF;
				break;
			}
			svm->param.C = svm->param.nu = svmC;
			svm->param.gamma = 1 / kernelGamma;
			svm->param.degree = kernelDegree;
		}
	}
}

Classifier *ClassSVM::GetClassifier()
{
	int svmType = params->svmTypeCombo->currentIndex();
	Classifier *classifier = 0;
	switch(svmType)
	{
	case 2:
		classifier = new ClassifierRVM();
		break;
	case 3:
		classifier = new ClassifierPegasos();
		break;
	default:
		classifier = new ClassifierSVM();
		break;
	}
	SetParams(classifier);
	return classifier;
}

void ClassSVM::DrawInfo(Canvas *canvas, Classifier *classifier)
{
	if(!canvas || !classifier) return;
	int w = canvas->width();
	int h = canvas->height();
	QPixmap infoPixmap(w, h);
	QBitmap bitmap(w,h);
	bitmap.clear();
	infoPixmap.setMask(bitmap);
	infoPixmap.fill(Qt::transparent);

	QPainter painter(&infoPixmap);
	painter.setRenderHint(QPainter::Antialiasing);
	if(classifier->type == CLASS_RVM || classifier->type == CLASS_PEG)
	{
		// we want to draw the support vectors
		vector<fvec> sv;
		if(classifier->type == CLASS_RVM) sv = ((ClassifierRVM*)classifier)->GetSVs();
		else sv = ((ClassifierPegasos*)classifier)->GetSVs();
		int radius = 9;
		FOR(i, sv.size())
		{
			QPoint point((u32)(sv[i][0]*w),(u32)(sv[i][1]*h));
			painter.setPen(QPen(Qt::black,6));
			painter.drawEllipse(point, radius, radius);
			painter.setPen(QPen(Qt::white,4));
			painter.drawEllipse(point, radius, radius);
		}
	}
	else if(classifier->type == CLASS_SVM)
	{
		// we want to draw the support vectors
		svm_model *svm = ((ClassifierSVM*)classifier)->GetModel();
		if(svm)
		{
			f32 sv[2];
			FOR(i, svm->l)
			{
				FOR(j, 2)
				{
					sv[j] = (f32)svm->SV[i][j].value;
				}
				int radius = 9;
				QPoint point((u32)(sv[0]*w),(u32)(sv[1]*h));
				if(abs((*svm->sv_coef)[i]) == svm->param.C)
				{
					painter.setPen(QPen(Qt::black, 6));
					painter.drawEllipse(point, radius, radius);
					painter.setPen(QPen(Qt::white,4));
					painter.drawEllipse(point, radius, radius);
				}
				else
				{
					painter.setPen(QPen(Qt::white, 6));
					painter.drawEllipse(point, radius, radius);
					painter.setPen(QPen(Qt::black,4));
					painter.drawEllipse(point, radius, radius);
				}
			}
		}
	}
	canvas->infoPixmap = infoPixmap;
}

void ClassSVM::Draw(Canvas *canvas, Classifier *classifier)
{
	if(!classifier || !canvas) return;
	canvas->liveTrajectory.clear();
	int w = canvas->width();
	int h = canvas->height();

	int posClass = 1;

	canvas->modelPixmap = QPixmap();

	DrawInfo(canvas, classifier);

	bool bUseMinMax = false;

	float resMin = FLT_MAX;
	float resMax = -FLT_MAX;
	if(bUseMinMax)
	{
		// TODO: get the min and max for all samples
		std::vector<fvec> samples = canvas->data->GetSamples();
		FOR(i, samples.size())
		{
			float val = classifier->Test(samples[i]);
			if(val > resMax) resMax = val;
			if(val < resMin) resMin = val;
		}
		if(resMin == resMax) resMin -= 3;
	}

	// we draw the samples
	canvas->modelPixmap = QPixmap(w,h);
	QBitmap bitmap(w,h);
	bitmap.clear();
	canvas->modelPixmap.setMask(bitmap);
	canvas->modelPixmap.fill(Qt::transparent);
	QPainter painter(&canvas->modelPixmap);
	painter.setRenderHint(QPainter::Antialiasing, true);
	FOR(i, canvas->data->GetCount())
	{
		fvec sample = canvas->data->GetSample(i);
		int label = canvas->data->GetLabel(i);
		QPointF point(canvas->data->GetSample(i)[0]*w,canvas->data->GetSample(i)[1]*h);
		float response = classifier->Test(sample);
		if(response > 0)
		{
			if(label == posClass) Canvas::drawSample(painter, point, 9, 2);
			else Canvas::drawCross(painter, point, 6, 2);
		}
		else
		{
			if(label != posClass) Canvas::drawSample(painter, point, 9, 0);
			else Canvas::drawCross(painter, point, 6, 0);
		}
	}

	canvas->repaint();
	canvas->confidencePixmap = QPixmap();
}

void ClassSVM::SaveOptions(QSettings &settings)
{
	settings.setValue("kernelDeg", params->kernelDegSpin->value());
	settings.setValue("kernelType", params->kernelTypeCombo->currentIndex());
	settings.setValue("kernelWidth", params->kernelWidthSpin->value());
	settings.setValue("svmC", params->svmCSpin->value());
	settings.setValue("svmType", params->svmTypeCombo->currentIndex());
}

bool ClassSVM::LoadOptions(QSettings &settings)
{
	if(settings.contains("kernelDeg")) params->kernelDegSpin->setValue(settings.value("kernelDeg").toFloat());
	if(settings.contains("kernelType")) params->kernelTypeCombo->setCurrentIndex(settings.value("kernelType").toInt());
	if(settings.contains("kernelWidth")) params->kernelWidthSpin->setValue(settings.value("kernelWidth").toFloat());
	if(settings.contains("svmC")) params->svmCSpin->setValue(settings.value("svmC").toFloat());
	if(settings.contains("svmType")) params->svmTypeCombo->setCurrentIndex(settings.value("svmType").toInt());
	return true;
}

void ClassSVM::SaveParams(std::ofstream &file)
{
	file << "classificationOptions" << ":" << "kernelDeg" << " " << params->kernelDegSpin->value() << std::endl;
	file << "classificationOptions" << ":" << "kernelType" << " " << params->kernelTypeCombo->currentIndex() << std::endl;
	file << "classificationOptions" << ":" << "kernelWidth" << " " << params->kernelWidthSpin->value() << std::endl;
	file << "classificationOptions" << ":" << "svmC" << " " << params->svmCSpin->value() << std::endl;
	file << "classificationOptions" << ":" << "svmType" << " " << params->svmTypeCombo->currentIndex() << std::endl;
}

bool ClassSVM::LoadParams(char *line, float value)
{
	if(endsWith(line,"kernelDeg")) params->kernelDegSpin->setValue((int)value);
	if(endsWith(line,"kernelType")) params->kernelTypeCombo->setCurrentIndex((int)value);
	if(endsWith(line,"kernelWidth")) params->kernelWidthSpin->setValue(value);
	if(endsWith(line,"svmC")) params->svmCSpin->setValue(value);
	if(endsWith(line,"svmType")) params->svmTypeCombo->setCurrentIndex((int)value);
	return true;
}
