/*********************************************************************
MLDemos: A User-Friendly visualization toolkit for machine learning
Copyright (C) 2010  Basilio Noris
Contact: mldemos@b4silio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License,
version 3 as published by the Free Software Foundation.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/
#include "interfaceSVMRegress.h"
#include <QPixmap>
#include <QBitmap>
#include <QPainter>

using namespace std;

RegrSVM::RegrSVM()
{
	params = new Ui::ParametersRegr();
	params->setupUi(widget = new QWidget());
	connect(params->svmTypeCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(ChangeOptions()));
}

void RegrSVM::ChangeOptions()
{
	params->svmCLabel->setText("C");
	params->svmPSpin->setRange(0.0001, 1.0);
	params->svmPSpin->setSingleStep(0.01);
	params->svmPSpin->setDecimals(4);
	params->svmCSpin->setEnabled(true);
	params->svmCSpin->setRange(0.1, 9999.9);
	params->svmCSpin->setDecimals(1);
	switch(params->svmTypeCombo->currentIndex())
	{
	case 0: // C-SVM
		params->svmEpsLabel->setText("eps");
		break;
	case 1: // Nu-SVM
		params->svmEpsLabel->setText("Nu");
		break;
	case 2: // RVM
		params->svmCSpin->setEnabled(false);
		params->svmEpsLabel->setText("eps");
		break;
	case 3: // SOGP
		params->svmEpsLabel->setText("Noise");
		params->svmCLabel->setText("Capacity");
		params->svmCSpin->setRange(-1, 500);
		params->svmCSpin->setDecimals(0);
		params->svmPSpin->setRange(0.001, 1.0);
		params->svmPSpin->setSingleStep(0.01);
		params->svmPSpin->setDecimals(3);
		break;
	case 4:
		params->svmEpsLabel->setText("Tolerance");
		params->svmCLabel->setText("Capacity");
		params->svmCSpin->setRange(0, 1000);
		params->svmCSpin->setDecimals(0);
		params->svmPSpin->setRange(0.0001, 1.0);
		params->svmPSpin->setSingleStep(0.001);
		params->svmPSpin->setDecimals(4);
		break;
	}
}

void RegrSVM::SetParams(Regressor *regressor)
{
	if(!regressor) return;
	int kernelMethod = params->svmTypeCombo->currentIndex();
	float svmC = params->svmCSpin->value();
	int kernelType = params->kernelTypeCombo->currentIndex();
	float kernelGamma = params->kernelWidthSpin->value();
	float kernelDegree = params->kernelDegSpin->value();
	float svmP = params->svmPSpin->value();

	if(kernelMethod == 2) // rvm
	{
		RegressorRVM *rvm = (RegressorRVM*)regressor;
		rvm->SetParams(svmP, kernelType, kernelGamma, kernelDegree);
	}
	else if(kernelMethod == 3) // sogp
	{
		RegressorGPR *gpr = (RegressorGPR*)regressor;
		int capacity = svmC;
		double kernelNoise = svmP;
		gpr->SetParams(kernelGamma, kernelNoise, capacity, kernelType, kernelDegree);
	}
	else if(kernelMethod == 4 ) // KRLS
	{
		RegressorKRLS *krls = (RegressorKRLS*)regressor;
		int capacity = svmC;
		double epsilon = svmP;
		krls->SetParams(epsilon, capacity, kernelType, kernelGamma, kernelDegree);
	}
	else
	{
		RegressorSVR *svm = (RegressorSVR*)regressor;
		switch(kernelMethod)
		{
		case 0:
			svm->param.svm_type = EPSILON_SVR;
			break;
		case 1:
			svm->param.svm_type = NU_SVR;
			break;
		}
		switch(kernelType)
		{
		case 0:
			svm->param.kernel_type = LINEAR;
			break;
		case 1:
			svm->param.kernel_type = POLY;
			break;
		case 2:
			svm->param.kernel_type = RBF;
			break;
		}
		svm->param.C = svmC;
		svm->param.nu = svmP;
		svm->param.p = svmP;
		svm->param.gamma = 1 / kernelGamma;
		svm->param.degree = kernelDegree;
	}
}

Regressor *RegrSVM::GetRegressor()
{
	int svmType = params->svmTypeCombo->currentIndex();
	Regressor *regressor = 0;
	switch(svmType)
	{
	case 2:
		regressor = new RegressorRVM();
		break;
	case 3:
		regressor = new RegressorGPR();
		break;
	case 4:
		regressor = new RegressorKRLS();
		break;
	default:
		regressor = new RegressorSVR();
		break;
	}
	SetParams(regressor);
	return regressor;
}

void DrawArrow( const QPointF &ppt, const QPointF &pt, double sze, QPainter &painter)
{
	QPointF pd, pa, pb;
	double tangent;

	pd = ppt - pt;
	if (pd.x() == 0 && pd.y() == 0)
		return;
	tangent = atan2 ((double) pd.y(), (double) pd.x());
	pa.setX(sze * cos (tangent + M_PI / 7.f) + pt.x());
	pa.setY(sze * sin (tangent + M_PI / 7.f) + pt.y());
	pb.setX(sze * cos (tangent - M_PI / 7.f) + pt.x());
	pb.setY(sze * sin (tangent - M_PI / 7.f) + pt.y());
	//-- connect the dots...
	painter.drawLine(pt, ppt);
	painter.drawLine(pt, pa);
	painter.drawLine(pt, pb);
}

void RegrSVM::DrawInfo(Canvas *canvas, Regressor *regressor)
{
	if(!canvas || !regressor) return;
	int w = canvas->width();
	int h = canvas->height();
	QPixmap infoPixmap(w, h);
	QBitmap bitmap(w,h);
	bitmap.clear();
	infoPixmap.setMask(bitmap);
	infoPixmap.fill(Qt::transparent);

	QPainter painter(&infoPixmap);
	painter.setRenderHint(QPainter::Antialiasing);

	if(regressor->type == REGR_RVM || regressor->type == REGR_KRLS)
	{
		vector<fvec> sv = (regressor->type == REGR_KRLS) ?
				((RegressorKRLS*)regressor)->GetSVs() :
				((RegressorRVM*)regressor)->GetSVs();
		int radius = 9;
		FOR(i, sv.size())
		{
			QPoint point((u32)(sv[i][0]*w),(u32)(sv[i][1]*h));
			painter.setPen(QPen(Qt::black,6));
			painter.drawEllipse(point, radius, radius);
			painter.setPen(QPen(Qt::white,3));
			painter.drawEllipse(point, radius, radius);
		}
	}
	else if(regressor->type == REGR_SVR)
	{
		// we want to draw the support vectors
		svm_model *svm = ((RegressorSVR*)regressor)->GetModel();
		if(svm)
		{
			std::vector<fvec> samples = canvas->data->GetSamples();
			fvec sv;
			sv.resize(2,0);
			FOR(i, svm->l)
			{
				sv[0] = (f32)svm->SV[i][0].value;
				FOR(j, samples.size())
				{
					if(sv[0] == samples[j][0])
					{
						sv[1] = samples[j][1];
						break;
					}
				}
				int radius = 7;
				QPoint point((u32)(sv[0]*w),(u32)(sv[1]*h));
				if(abs((*svm->sv_coef)[i]) == svm->param.C)
				{
					painter.setPen(QPen(Qt::black, 4));
					painter.drawEllipse(point, radius, radius);
					painter.setPen(Qt::white);
					painter.drawEllipse(point, radius, radius);
				}
				else
				{
					painter.setPen(Qt::black);
					painter.drawEllipse(point, radius, radius);
				}
			}
		}
	}
	else if(regressor->type == REGR_GPR)
	{
		RegressorGPR * gpr = (RegressorGPR*)regressor;
		int radius = 8;
		FOR(i, gpr->GetBasisCount())
		{
			fvec basis = gpr->GetBasisVector(i);
			fvec res = gpr->Test(basis);
			QPoint pt1((u32)(basis[0]*w),(u32)(res[0]*h));
			QPoint pt2 = pt1 + QPoint(0,(basis[1]>0 ? 1 : -1)*radius);
			QPoint pt3 = pt2 + QPoint(0,(basis[1]>0 ? 1 : -1)*50);
			painter.setPen(QPen(Qt::red,3));
			painter.drawEllipse(pt1, radius, radius);
			painter.setPen(QPen(Qt::red,min(4.f,max(fabs(basis[1])/5,0.5f))));
			DrawArrow(pt2,pt3,10,painter);
		}
	}
	canvas->infoPixmap = infoPixmap;
}

void RegrSVM::Draw(Canvas *canvas, Regressor *regressor)
{
	if(!regressor || !canvas) return;
	canvas->liveTrajectory.clear();
	DrawInfo(canvas, regressor);
	int w = canvas->width();
	int h = canvas->height();
#define RES 256
	canvas->confidencePixmap = QPixmap(w,h);
	canvas->modelPixmap = QPixmap(w,h);
	QBitmap bitmap(w,h);
	bitmap.clear();
	canvas->modelPixmap.setMask(bitmap);
	canvas->modelPixmap.fill(Qt::transparent);
	QPainter painter(&canvas->modelPixmap);
	painter.setRenderHint(QPainter::Antialiasing, true);

	fvec sample;
	sample.resize(2,0);
	if(regressor->type == REGR_KRLS)
	{
		canvas->confidencePixmap = QPixmap();
		int steps = RES;
		QPointF oldPoint(-FLT_MAX,-FLT_MAX);
		FOR(x, steps)
		{
			sample[0] = (x/(float)steps);
			fvec res = regressor->Test(sample);
			if(res[0] != res[0]) continue;
			QPointF point(sample[0]*w, res[0]*h);
			if(x)
			{
				painter.setPen(QPen(Qt::black, 1));
				painter.drawLine(point, oldPoint);
				painter.setPen(QPen(Qt::black, 0.5));
				//				painter.drawLine(point+QPointF(0,eps*h), oldPoint+QPointF(0,eps*h));
				//				painter.drawLine(point-QPointF(0,eps*h), oldPoint-QPointF(0,eps*h));
			}
			oldPoint = point;
		}
	}
	if(regressor->type == REGR_RVM)
	{
		canvas->confidencePixmap = QPixmap();
		int steps = RES;
		QPointF oldPoint(-FLT_MAX,-FLT_MAX);
		FOR(x, steps)
		{
			sample[0] = (x/(float)steps);
			fvec res = regressor->Test(sample);
			if(res[0] != res[0]) continue;
			QPointF point(sample[0]*w, res[0]*h);
			if(x)
			{
				painter.setPen(QPen(Qt::black, 1));
				painter.drawLine(point, oldPoint);
				painter.setPen(QPen(Qt::black, 0.5));
//				painter.drawLine(point+QPointF(0,eps*h), oldPoint+QPointF(0,eps*h));
//				painter.drawLine(point-QPointF(0,eps*h), oldPoint-QPointF(0,eps*h));
			}
			oldPoint = point;
		}
	}
	else if(regressor->type == REGR_SVR)
	{
		canvas->confidencePixmap = QPixmap();
		svm_parameter params = ((RegressorSVR *)regressor)->param;

		float eps = params.p;
		if(params.svm_type == NU_SVR) eps = ((RegressorSVR *)regressor)->GetModel()->eps[0];
		int steps = RES;
		QPointF oldPoint(-FLT_MAX,-FLT_MAX);
		FOR(x, steps)
		{
			sample[0] = (x/(float)steps);
			fvec res = regressor->Test(sample);
			if(res[0] != res[0]) continue;
			QPointF point(sample[0]*w, res[0]*h);
			if(x)
			{
				painter.setPen(QPen(Qt::black, 1));
				painter.drawLine(point, oldPoint);
				painter.setPen(QPen(Qt::black, 0.5));
				painter.drawLine(point+QPointF(0,eps*h), oldPoint+QPointF(0,eps*h));
				painter.drawLine(point-QPointF(0,eps*h), oldPoint-QPointF(0,eps*h));
			}
			oldPoint = point;
		}
	}
	else if(regressor->type == REGR_GPR)
	{
		IplImage *image = cvCreateImage(cvSize(w,h), 8, 3);
		cvSet(image, CV_RGB(255,255,255));
		((RegressorGPR *)regressor)->Draw(image);
		canvas->confidencePixmap = Canvas::toPixmap(image);
		IMKILL(image);
		int steps = RES;
		QPointF oldPoint(-FLT_MAX,-FLT_MAX);
		QPointF oldPointUp(-FLT_MAX,-FLT_MAX);
		QPointF oldPointDown(-FLT_MAX,-FLT_MAX);
		FOR(x, steps)
		{
			sample[0] = (x/(float)steps);
			fvec res = regressor->Test(sample);
			if(res[0] != res[0] || res[1] != res[1]) continue;
			QPointF point(sample[0]*w, res[0]*h);
			QPointF pointUp = point + QPointF(0,res[1]*h);
			QPointF pointDown = point - QPointF(0,res[1]*h);
			if(x)
			{
				painter.setPen(QPen(Qt::black, 1));
				painter.drawLine(point, oldPoint);
				painter.setPen(QPen(Qt::black, 0.5));
				painter.drawLine(pointUp, oldPointUp);
				painter.drawLine(pointDown, oldPointDown);
			}
			oldPoint = point;
			oldPointUp = pointUp;
			oldPointDown = pointDown;
		}
	}

	canvas->repaint();
}

void RegrSVM::SaveOptions(QSettings &settings)
{
	settings.setValue("kernelDeg", params->kernelDegSpin->value());
	settings.setValue("kernelType", params->kernelTypeCombo->currentIndex());
	settings.setValue("kernelWidth", params->kernelWidthSpin->value());
	settings.setValue("svmC", params->svmCSpin->value());
	settings.setValue("svmP", params->svmPSpin->value());
	settings.setValue("svmType", params->svmTypeCombo->currentIndex());
}

bool RegrSVM::LoadOptions(QSettings &settings)
{
	if(settings.contains("kernelDeg")) params->kernelDegSpin->setValue(settings.value("kernelDeg").toFloat());
	if(settings.contains("kernelType")) params->kernelTypeCombo->setCurrentIndex(settings.value("kernelType").toInt());
	if(settings.contains("kernelWidth")) params->kernelWidthSpin->setValue(settings.value("kernelWidth").toFloat());
	if(settings.contains("svmC")) params->svmCSpin->setValue(settings.value("svmC").toFloat());
	if(settings.contains("svmP")) params->svmPSpin->setValue(settings.value("svmP").toFloat());
	if(settings.contains("svmType")) params->svmTypeCombo->setCurrentIndex(settings.value("svmType").toInt());
	return true;
}

void RegrSVM::SaveParams(std::ofstream &file)
{
	file << "regressionOptions" << ":" << "kernelDeg" << " " << params->kernelDegSpin->value() << std::endl;
	file << "regressionOptions" << ":" << "kernelType" << " " << params->kernelTypeCombo->currentIndex() << std::endl;
	file << "regressionOptions" << ":" << "kernelWidth" << " " << params->kernelWidthSpin->value() << std::endl;
	file << "regressionOptions" << ":" << "svmC" << " " << params->svmCSpin->value() << std::endl;
	file << "regressionOptions" << ":" << "svmP" << " " << params->svmPSpin->value() << std::endl;
	file << "regressionOptions" << ":" << "svmType" << " " << params->svmTypeCombo->currentIndex() << std::endl;
}

bool RegrSVM::LoadParams(char *line, float value)
{
	if(endsWith(line,"kernelDeg")) params->kernelDegSpin->setValue((int)value);
	if(endsWith(line,"kernelType")) params->kernelTypeCombo->setCurrentIndex((int)value);
	if(endsWith(line,"kernelWidth")) params->kernelWidthSpin->setValue(value);
	if(endsWith(line,"svmC")) params->svmCSpin->setValue(value);
	if(endsWith(line,"svmP")) params->svmPSpin->setValue(value);
	if(endsWith(line,"svmType")) params->svmTypeCombo->setCurrentIndex((int)value);
	return true;
}
