/*********************************************************************
MLDemos: A User-Friendly visualization toolkit for machine learning
Copyright (C) 2010  Basilio Noris
Contact: mldemos@b4silio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/
#include <QtGui>
#include <QWidget>
#include <QSize>
#include <QPixmap>
#include <QPainter>
#include <QPen>
#include <QImage>

#include "public.h"
#include "canvas.h"

using namespace std;

//DatasetManager Canvas::data;
bool Canvas::bCrossesAsDots = true;

Canvas::Canvas(QWidget *parent)
: QWidget(parent),
bDisplayMap(false),
bDisplayInfo(false),
bDisplaySingle(false),
bDisplaySamples(true),
bDisplayTrajectories(true),
bDisplayLearned(true),
crosshair(QPainterPath()),
bShowCrosshair(false),
bNewCrosshair(true),
trajectoryCenterType(0),
trajectoryResampleType(1),
trajectoryResampleCount(100),
liveTrajectory(vector<fvec>()),
centers(map<int,fvec>()),
drawnSamples(0),
drawnTrajectories(0),
bDrawing(false),
data(new DatasetManager())
{

	resize(640,480);

	setMouseTracking(true);
	setCursor(Qt::CrossCursor);
	setBackgroundRole(QPalette::Base);
//	setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
//	setFixedSize(640,480);
	setMouseTracking(true);

	QPalette p(palette());
	p.setColor(backgroundRole(), Qt::white);
	setPalette(p);
	show();
}

Canvas::~Canvas()
{
	if(data) DEL(data);
}

void Canvas::SetConfidenceMap(QImage image)
{
	confidencePixmap = QPixmap::fromImage(image);
	repaint();
}

void Canvas::SetModelImage(QImage image)
{
	modelPixmap = QPixmap::fromImage(image);
	repaint();
}

void Canvas::paintEvent(QPaintEvent *event)
{
	if(bDrawing) return;
	bDrawing = true;
	QPainter painter(this);
	painter.setBackgroundMode(Qt::OpaqueMode);
	painter.setBackground(Qt::white);

	painter.fillRect(geometry(),Qt::white);

	if(bDisplayMap && !confidencePixmap.isNull()) painter.drawPixmap(geometry(), confidencePixmap);
	painter.setRenderHint(QPainter::Antialiasing);
	painter.setRenderHint(QPainter::HighQualityAntialiasing);

	if(bDisplaySamples)
	{
		DrawSamples();
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), samplesPixmap);
	}
	if(bDisplayTrajectories)
	{
		DrawTrajectories();
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), trajectoriesPixmap);
	}
	if(bDisplayLearned && !modelPixmap.isNull())
	{
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), modelPixmap);
	}
	if(bDisplayInfo && !infoPixmap.isNull())
	{
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), infoPixmap);
	}
	if(bShowCrosshair)
	{
		if(bNewCrosshair) emit DrawCrosshair();
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPath(crosshair.translated(mouse));
		if(liveTrajectory.size()) DrawLiveTrajectory(painter);
	}
	bDrawing = false;
}

void Canvas::DrawSamples()
{
	int radius = 10;
	if(!data->GetCount())
	{
		int w = width();
		int h = height();
		samplesPixmap = QPixmap(w,h);
		QBitmap bitmap(w,h);
		bitmap.clear();
		samplesPixmap.setMask(bitmap);
		samplesPixmap.fill(Qt::transparent);
		drawnSamples = 0;
		return;
	}
	if(drawnSamples == data->GetCount()) return;
	if(drawnSamples > data->GetCount()) drawnSamples = 0;

	if(!drawnSamples || samplesPixmap.isNull())
	{
		int w = width();
		int h = height();
		samplesPixmap = QPixmap(w,h);
		QBitmap bitmap(w,h);
		bitmap.clear();
		samplesPixmap.setMask(bitmap);
		samplesPixmap.fill(Qt::transparent);
		drawnSamples = 0;
	}
	QPainter painter(&samplesPixmap);
	painter.setRenderHint(QPainter::Antialiasing, true);
	painter.setRenderHint(QPainter::HighQualityAntialiasing);
	for(int i=drawnSamples; i<data->GetCount(); i++)
	{
		if(data->GetFlag(i) == _TRAJ) continue;
		fvec sample = data->GetSample(i);
		int x = int(sample[0]*geometry().width());
		int y = int(sample[1]*geometry().height());
		int label = data->GetLabel(i);
		Canvas::drawSample(painter, QPoint(x,y), (data->GetFlag(i)==_TRAJ)?5:radius, bDisplaySingle ? 0 : label);
	}
	drawnSamples = data->GetCount();
}

void Canvas::DrawTrajectories()
{
	int w = width();
	int h = height();
	int count = data->GetCount();
	if(!count || (!data->GetSequences().size() && (data->GetFlag(count-1) != _TRAJ)))
	{
		trajectoriesPixmap = QPixmap(w,h);
		QBitmap bitmap(w,h);
		bitmap.clear();
		trajectoriesPixmap.setMask(bitmap);
		trajectoriesPixmap.fill(Qt::transparent);
		drawnTrajectories = 0;
	}

	bool bDrawing = false;

	vector<ipair> sequences = data->GetSequences();
	int start=0, stop=0;
	if(data->GetFlag(count-1) == _TRAJ)
	{
		if(sequences.size()) stop = sequences[sequences.size()-1].second;
		if(stop < count-1) // there's an unfinished trajectory
		{
			stop++;
			for(start=count-1; start >= stop && data->GetFlag(start) == _TRAJ; start--);
			sequences.push_back(ipair(start+(sequences.size() ? 1 : 0),count-1));
			bDrawing = true;
		}
	}
	if(!bDrawing && drawnTrajectories == sequences.size()) return;
	if(drawnTrajectories > sequences.size()) drawnTrajectories = 0;

	if(!drawnTrajectories || trajectoriesPixmap.isNull())
	{
		trajectoriesPixmap = QPixmap(w,h);
		QBitmap bitmap(w,h);
		bitmap.clear();
		trajectoriesPixmap.setMask(bitmap);
		trajectoriesPixmap.fill(Qt::transparent);
		drawnTrajectories = 0;
	}

	QPainter painter(&trajectoriesPixmap);
	painter.setRenderHint(QPainter::Antialiasing);
	painter.setRenderHint(QPainter::HighQualityAntialiasing);

	vector<fvec> samples = data->GetSamples();

	map<int,int> counts;
	centers.clear();
	if(trajectoryCenterType)
	{
		FOR(i, sequences.size())
		{
			int index = sequences[i].first;
			if(trajectoryCenterType==1) // end
			{
				index = sequences[i].second;
			}
			int label = data->GetLabel(index);
			if(!centers.count(label))
			{
				fvec center;
				center.resize(2,0);
				centers[label] = center;
				counts[label] = 0;
			}
			centers[label] += samples[index];
			counts[label]++;
		}

		for(map<int,int>::iterator p = counts.begin(); p!=counts.end(); ++p)
		{
			int label = p->first;
			centers[label] /= p->second;
		}
	}

	vector< vector<fvec> > trajectories;
	vector<fvec> diffs;
	ivec trajLabels;
	for(int i=drawnTrajectories; i< sequences.size(); i++)
	{
		start = sequences[i].first;
		stop = sequences[i].second;
		int label = data->GetLabel(start);
		fvec diff;
		if(trajectoryCenterType && (i < sequences.size()-1 || !bDrawing))
		{
			diff = centers[label] - samples[trajectoryCenterType==1?stop:start];
		}
		else diff.resize(2,0);
		vector<fvec> trajectory;
		trajectory.resize(stop-start+1);
		int pos = 0;
		for (int j=start; j<=stop; j++)
		{
			trajectory[pos++] = samples[j] + diff;
		}
		switch (trajectoryResampleType)
		{
		case 0: // do nothing
			break;
		case 1: // uniform resampling
			{
				if(i < sequences.size()-1 || !bDrawing)
				{
					trajectory = interpolate(trajectory, trajectoryResampleCount);
				}
			}
			break;
		case 2: // spline resampling
			{
				if(i < sequences.size()-1 || !bDrawing)
				{
					trajectory = interpolate(trajectory, trajectoryResampleCount);
				}
			}
			break;
		}
		trajectories.push_back(trajectory);
		trajLabels.push_back(data->GetLabel(start));
	}

	// let's draw the trajectories
	FOR(i, trajectories.size())
	{
		fvec oldPt = trajectories[i][0];
		int count = trajectories[i].size();
		int label = trajLabels[i];
		FOR(j, count-1)
		{
			fvec pt = trajectories[i][j+1];
			painter.setPen(QPen(Qt::black, 0.5));
			painter.drawLine(pt[0]*w, pt[1]*h, oldPt[0]*w, oldPt[1]*h);
			if(j<count-2) Canvas::drawSample(painter, QPoint(pt[0]*w, pt[1]*h), 5, bDisplaySingle ? 0 : label);
			oldPt = pt;
		}
		painter.setBrush(Qt::NoBrush);
		painter.setPen(Qt::green);
		painter.drawEllipse(QPointF(trajectories[i][0][0]*w, (float)trajectories[i][0][1]*h), 5, 5);
		if(!bDrawing)
		{
			painter.setPen(Qt::red);
			painter.drawEllipse(QPointF(trajectories[i][count-1][0]*w, trajectories[i][count-1][1]*h), 5, 5);
		}
	}
	drawnTrajectories = !bDrawing ? sequences.size() : sequences.size()-1;
}

void Canvas::DrawLiveTrajectory(QPainter &painter)
{
	if(!liveTrajectory.size() || !liveTrajectory[0].size()) return;
	int w = width();
	int h = height();
	fvec oldPt = liveTrajectory[0];
	int count = liveTrajectory.size();
	FOR(j, count-1)
	{
		fvec pt = liveTrajectory[j+1];
		if(!pt.size()) break;
		int label = 1;
		if(bDisplayMap)
		{
			painter.setPen(QPen(Qt::white, 3));
			painter.drawLine(pt[0]*w, pt[1]*h, oldPt[0]*w, oldPt[1]*h);
			painter.setPen(QPen(Qt::black, 1));
			painter.drawLine(pt[0]*w, pt[1]*h, oldPt[0]*w, oldPt[1]*h);
		}
		else
		{
			painter.setPen(QPen(Qt::magenta, 2));
			painter.drawLine(pt[0]*w, pt[1]*h, oldPt[0]*w, oldPt[1]*h);
		}
		//if(j<count-2) Canvas::drawSample(painter, QPoint(pt[0]*w, pt[1]*h), 5, label);
		oldPt = pt;
	}
	painter.setBrush(Qt::NoBrush);
	painter.setPen(Qt::green);
	painter.drawEllipse(QPointF(liveTrajectory[0][0]*w, (float)liveTrajectory[0][1]*h), 5, 5);
	painter.setPen(Qt::red);
	painter.drawEllipse(QPointF(liveTrajectory[count-1][0]*w, liveTrajectory[count-1][1]*h), 5, 5);
}

void Canvas::resizeEvent( QResizeEvent *event )
{
	bNewCrosshair = true;
}

void Canvas::mousePressEvent( QMouseEvent *event )
{
	int x = event->x();
	int y = event->y();

	fvec sample;
	sample.push_back(x/(double)geometry().width());
	sample.push_back(y/(double)geometry().height());

	int label = 0;
	if(event->button()==Qt::LeftButton) label = 1;
	if(event->button()==Qt::RightButton) label = 0;

	emit Drawing(sample, label);
}

void Canvas::mouseReleaseEvent( QMouseEvent *event )
{
	int x = event->x();
	int y = event->y();

	fvec sample;
	sample.push_back(x/(double)geometry().width());
	sample.push_back(y/(double)geometry().height());

	int label = 0;
	if(event->button()==Qt::LeftButton) label = 1;
	if(event->button()==Qt::RightButton) label = 0;

	//emit Drawing(sample, label);
	emit Released();
}

void Canvas::enterEvent(QEvent *event)
{
	bShowCrosshair = true;
	repaint();
}

void Canvas::focusOutEvent(QFocusEvent *event)
{
	bShowCrosshair = false;
	bNewCrosshair = true;
	repaint();
}

void Canvas::leaveEvent(QEvent *event)
{
	bShowCrosshair = false;
	bNewCrosshair = true;
	repaint();
}

void Canvas::mouseMoveEvent( QMouseEvent *event )
{
	int x = event->x();
	int y = event->y();
	mouse = QPoint(x,y);
	fvec sample;
	sample.push_back(x/(double)geometry().width());
	sample.push_back(y/(double)geometry().height());
	if(x < 0 || x > width() || y < 0 || y < height());

	if(event->buttons() != Qt::LeftButton && event->buttons() != Qt::RightButton )
	{
		emit Navigation(sample);
		repaint();
	}
	else
	{
		int label = 0;
		if(event->buttons()==Qt::LeftButton) label = 1;
		if(event->buttons()==Qt::RightButton) label = 0;
		emit Drawing(sample, label);
	}
}

bool Canvas::SaveScreenshot( QString filename )
{
	QPixmap screenshot = GetScreenshot();
	return screenshot.save(filename);
}

QPixmap Canvas::GetScreenshot()
{
	QPixmap screenshot(width(), height());
	QPainter painter(&screenshot);
	painter.setBackgroundMode(Qt::OpaqueMode);
	painter.setBackground(Qt::white);

	painter.fillRect(geometry(),Qt::white);

	if(bDisplayMap && !confidencePixmap.isNull()) painter.drawPixmap(geometry(), confidencePixmap);
	painter.setRenderHint(QPainter::Antialiasing);
	painter.setRenderHint(QPainter::HighQualityAntialiasing);

	if(bDisplaySamples)
	{
		DrawSamples();
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), samplesPixmap);
	}

	if(bDisplayTrajectories)
	{
		DrawTrajectories();
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), trajectoriesPixmap);
	}

	if(bDisplayLearned && !modelPixmap.isNull())
	{
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), modelPixmap);
	}


	if(bDisplayInfo && !infoPixmap.isNull())
	{
		painter.setBackgroundMode(Qt::TransparentMode);
		painter.drawPixmap(geometry(), infoPixmap);
	}
	return screenshot;
}


void Canvas::DeleteData( QPointF center, float radius )
{
	FOR(i, data->GetCount())
	{
		QPointF dataPoint(data->GetSample(i)[0],data->GetSample(i)[1]);
		QPoint dPoint((int)(dataPoint.x()*width()), (int)(dataPoint.y()*height()));
		QPointF point = this->mapToParent(dPoint);
		point -= center;
		if(sqrt(point.x()*point.x() + point.y()*point.y()) < radius)
		{
			data->RemoveSample(i);
			i--;
		}
	}
}
