/*********************************************************************
MLDemos: A User-Friendly visualization toolkit for machine learning
Copyright (C) 2010  Basilio Noris
Contact: mldemos@b4silio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/
#include "mldemos.h"
#include <QDebug>
#include <QtGui>
#include <fstream>
#include <QPixmap>
#include <QUrl>
#include <QBitmap>
#include <QSettings>
#include <QFileDialog>
#include "basicMath.h"
#include "fgmm/gaussian.h"


MLDemos::MLDemos(QString filename, QWidget *parent, Qt::WFlags flags)
    : QMainWindow(parent, flags), canvas(0), classifier(0), regressor(0), dynamical(0), clusterer(0), bIsRocNew(true), bIsCrossNew(true), trajectory(ipair(-1,-1))
{
    QApplication::setWindowIcon(QIcon(":/MLDemos/logo.png"));
    ui.setupUi(this);
    setAcceptDrops(true);

    connect(ui.actionExit, SIGNAL(triggered()), qApp, SLOT(quit()));
    connect(ui.actionAbout, SIGNAL(triggered()), this, SLOT(ShowAbout()));
    connect(ui.actionClearData, SIGNAL(triggered()), this, SLOT(ClearData()));
    connect(ui.actionClearModel, SIGNAL(triggered()), this, SLOT(Clear()));
    connect(ui.actionNew, SIGNAL(triggered()), this, SLOT(ClearData()));
    connect(ui.actionSave, SIGNAL(triggered()), this, SLOT(SaveData()));
    connect(ui.actionLoad, SIGNAL(triggered()), this, SLOT(LoadData()));
    connect(ui.actionExportOutput, SIGNAL(triggered()), this, SLOT(ExportOutput()));
    connect(ui.actionExportAnimation, SIGNAL(triggered()), this, SLOT(ExportAnimation()));

    initDialogs();
    initToolBars();
    initPlugins();

    LoadLayoutOptions();
    DisplayOptionChanged();
    UpdateInfo();
    drawTime.start();
    if(filename != "") Load(filename);
}

void MLDemos::initToolBars()
{
    actionNew = new QAction(QIcon(":/MLDemos/icons/new.png"), tr("&New"), this);
    actionNew->setShortcut(QKeySequence(tr("Ctrl+N")));
    actionNew->setStatusTip(tr("Clear Everything"));

    actionSave = new QAction(QIcon(":/MLDemos/icons/save.png"), tr("&Save"), this);
    actionSave->setShortcut(QKeySequence(tr("Ctrl+S")));
    actionSave->setStatusTip(tr("Save Data"));

    actionLoad = new QAction(QIcon(":/MLDemos/icons/load.png"), tr("&Load"), this);
    actionLoad->setShortcut(QKeySequence(tr("Ctrl+L")));
    actionLoad->setStatusTip(tr("Load Data"));

    actionClassifiers = new QAction(QIcon(":/MLDemos/icons/classify.png"), tr("&Classification"), this);
    actionClassifiers->setShortcut(QKeySequence(tr("C")));
    actionClassifiers->setStatusTip(tr("Classify the data"));
    actionClassifiers->setCheckable(true);

    actionRegression = new QAction(QIcon(":/MLDemos/icons/regress.png"), tr("&Regression"), this);
    actionRegression->setShortcut(QKeySequence(tr("R")));
    actionRegression->setStatusTip(tr("Estimate Regression"));
    actionRegression->setCheckable(true);

    actionDynamical = new QAction(QIcon(":/MLDemos/icons/dynamical.png"), tr("&Dynamical"), this);
    actionDynamical->setShortcut(QKeySequence(tr("D")));
    actionDynamical->setStatusTip(tr("Estimate Dynamical System"));
    actionDynamical->setCheckable(true);

    actionClustering = new QAction(QIcon(":/MLDemos/icons/cluster.png"), tr("C&lustering"), this);
    actionClustering->setShortcut(QKeySequence(tr("L")));
    actionClustering->setStatusTip(tr("Cluster the data"));
    actionClustering->setCheckable(true);

    actionDrawSamples = new QAction(QIcon(":/MLDemos/icons/draw.png"), tr("&Drawing"), this);
    actionDrawSamples->setShortcut(QKeySequence(tr("W")));
    actionDrawSamples->setStatusTip(tr("Show Sample Drawing Options"));
    actionDrawSamples->setCheckable(true);

    actionClearModel = new QAction(QIcon(":/MLDemos/icons/clearmodel.png"), tr("Clear Model"), this);
    actionClearModel->setShortcut(QKeySequence(tr("Shift+X")));
    actionClearModel->setStatusTip(tr("Clear current model"));

    actionClearData = new QAction(QIcon(":/MLDemos/icons/cleardata.png"), tr("Clear Data"), this);
    actionClearData->setShortcut(QKeySequence(tr("X")));
    actionClearData->setStatusTip(tr("Clear all data"));

    actionScreenshot = new QAction(QIcon(":/MLDemos/icons/screenshot.png"), tr("Save Screenshot"), this);
    actionScreenshot->setShortcut(QKeySequence(tr("Alt+S")));
    actionScreenshot->setStatusTip(tr("Save the current image to disk"));

    actionDisplayOptions = new QAction(QIcon(":/MLDemos/icons/display.png"), tr("Display &Options"), this);
    actionDisplayOptions->setShortcut(QKeySequence(tr("O")));
    actionDisplayOptions->setStatusTip(tr("Show Display Options"));
    actionDisplayOptions->setCheckable(true);

    actionShowStats = new QAction(QIcon(":/MLDemos/icons/stats.png"), tr("Data Info/Statistics"), this);
    actionShowStats->setShortcut(QKeySequence(tr("I")));
    actionShowStats->setStatusTip(tr("Display Data Information and Statistics"));
    actionShowStats->setCheckable(true);

    connect(actionClassifiers, SIGNAL(triggered()), this, SLOT(ShowOptionClass()));
    connect(actionRegression, SIGNAL(triggered()), this, SLOT(ShowOptionRegress()));
    connect(actionDynamical, SIGNAL(triggered()), this, SLOT(ShowOptionDynamical()));
    connect(actionClustering, SIGNAL(triggered()), this, SLOT(ShowOptionCluster()));
    connect(actionDrawSamples, SIGNAL(triggered()), this, SLOT(ShowSampleDrawing()));
    connect(actionDisplayOptions, SIGNAL(triggered()), this, SLOT(ShowOptionDisplay()));
    connect(actionClearData, SIGNAL(triggered()), this, SLOT(ClearData()));
    connect(actionClearModel, SIGNAL(triggered()), this, SLOT(Clear()));
    connect(actionScreenshot, SIGNAL(triggered()), this, SLOT(Screenshot()));
    connect(actionNew, SIGNAL(triggered()), this, SLOT(ClearData()));
    connect(actionSave, SIGNAL(triggered()), this, SLOT(SaveData()));
    connect(actionLoad, SIGNAL(triggered()), this, SLOT(LoadData()));
    connect(actionShowStats, SIGNAL(triggered()), this, SLOT(ShowStatsDialog()));

    /*
 connect(actionClearData, SIGNAL(triggered()), this, SLOT(ClearData()));
 connect(actionClearModel, SIGNAL(triggered()), this, SLOT(Clear()));
 connect(actionNew, SIGNAL(triggered()), this, SLOT(ClearData()));
 connect(actionSave, SIGNAL(triggered()), this, SLOT(SaveData()));
 connect(actionLoad, SIGNAL(triggered()), this, SLOT(LoadData()));
 */

    QToolBar *toolBar = addToolBar("Tools");
    toolBar->setObjectName("MainToolBar");
    toolBar->setMovable(false);
    toolBar->setFloatable(false);
    toolBar->setIconSize(QSize(64,64));
    toolBar->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);

    toolBar->addAction(actionNew);
    toolBar->addAction(actionLoad);
    toolBar->addAction(actionSave);
    toolBar->addSeparator();
    toolBar->addAction(actionClustering);
    toolBar->addAction(actionClassifiers);
    toolBar->addAction(actionRegression);
    toolBar->addAction(actionDynamical);
    toolBar->addSeparator();
    toolBar->addAction(actionClearModel);
    toolBar->addAction(actionClearData);
    toolBar->addSeparator();
    toolBar->addAction(actionDrawSamples);
    toolBar->addSeparator();
    toolBar->addAction(actionScreenshot);
    toolBar->addAction(actionDisplayOptions);
    toolBar->addAction(actionShowStats);
}

void MLDemos::initDialogs()
{

    /*
 classifierOptions = new Ui::ClassifierOptions();
 QFrame *classWidget = new QFrame();
 classifierOptions->setupUi(classWidget);
 classWidget->show();
 /**/

    drawToolbar = new Ui::DrawingToolbar();
    drawToolbarContext1 = new Ui::DrawingToolbarContext1();
    drawToolbarContext2 = new Ui::DrawingToolbarContext2();

    drawToolbar->setupUi(drawToolbarWidget = new QWidget());
    drawToolbarContext1->setupUi(drawContext1Widget = new QWidget());
    drawToolbarContext2->setupUi(drawContext2Widget = new QWidget());

    connect(qApp, SIGNAL(focusChanged(QWidget *,QWidget *)),this,SLOT(FocusChanged(QWidget *,QWidget *)));

    drawToolbarWidget->setWindowFlags(Qt::Tool);
    drawContext1Widget->setWindowFlags(Qt::FramelessWindowHint);
    drawContext2Widget->setWindowFlags(Qt::FramelessWindowHint);
    drawToolbarWidget->setFixedSize(drawToolbarWidget->size());

    drawToolbar->sprayButton->setContextMenuPolicy(Qt::CustomContextMenu);
    drawToolbar->ellipseButton->setContextMenuPolicy(Qt::CustomContextMenu);
    drawToolbar->lineButton->setContextMenuPolicy(Qt::CustomContextMenu);
    drawToolbar->eraseButton->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(drawToolbar->sprayButton, SIGNAL(customContextMenuRequested( const QPoint& )), this, SLOT(ShowContextMenuSpray(const QPoint &)));
    connect(drawToolbar->ellipseButton, SIGNAL(customContextMenuRequested( const QPoint& )), this, SLOT(ShowContextMenuEllipse(const QPoint &)));
    connect(drawToolbar->lineButton, SIGNAL(customContextMenuRequested( const QPoint& )), this, SLOT(ShowContextMenuLine(const QPoint &)));
    connect(drawToolbar->eraseButton, SIGNAL(customContextMenuRequested( const QPoint& )), this, SLOT(ShowContextMenuErase(const QPoint &)));

    displayOptions = new Ui::viewOptionDialog();
    aboutPanel = new Ui::aboutDialog();
    showStats = new Ui::statisticsDialog();

    optionsClassify = new Ui::optionsClassifyDialog();
    optionsClassify->setupUi(classifyDialog = new QDialog());
    optionsCluster = new Ui::optionsClusterDialog();
    optionsCluster->setupUi(clusterDialog = new QDialog());
    optionsRegress = new Ui::optionsRegressDialog();
    optionsRegress->setupUi(regressDialog = new QDialog());
    optionsDynamic = new Ui::optionsDynamicDialog();
    optionsDynamic->setupUi(dynamicDialog = new QDialog());

    displayOptions->setupUi(displayDialog = new QDialog());
    aboutPanel->setupUi(about = new QDialog());
    showStats->setupUi(statsDialog = new QDialog());

    QRect geo = showStats->rocWidget->geometry();
    //	qDebug() << geo;
    rocWidget = new QNamedWindow("ROC Curve", false, showStats->rocWidget);
    crossvalidWidget = new QNamedWindow("Cross Validation", false, showStats->crossvalidWidget);
    infoWidget = new QNamedWindow("Info", false, showStats->informationWidget);

    connect(showStats->tabWidget, SIGNAL(currentChanged(int)), this, SLOT(StatsChanged()));
    connect(rocWidget, SIGNAL(ResizeEvent(QResizeEvent *)), this, SLOT(StatsChanged()));
    connect(crossvalidWidget, SIGNAL(ResizeEvent(QResizeEvent *)), this, SLOT(StatsChanged()));
    connect(infoWidget, SIGNAL(ResizeEvent(QResizeEvent *)), this, SLOT(StatsChanged()));

    connect(classifyDialog, SIGNAL(rejected()), this, SLOT(HideOptionClass()));
    connect(clusterDialog, SIGNAL(rejected()), this, SLOT(HideOptionCluster()));
    connect(regressDialog, SIGNAL(rejected()), this, SLOT(HideOptionRegress()));
    connect(dynamicDialog, SIGNAL(rejected()), this, SLOT(HideOptionDynamical()));
    connect(displayDialog, SIGNAL(rejected()), this, SLOT(HideOptionDisplay()));
    connect(statsDialog, SIGNAL(rejected()), this, SLOT(HideStatsDialog()));

    connect(optionsClassify->classifyButton, SIGNAL(clicked()), this, SLOT(Classify()));
    connect(optionsClassify->clearButton, SIGNAL(clicked()), this, SLOT(Clear()));
    connect(optionsClassify->rocButton, SIGNAL(clicked()), this, SLOT(ShowRoc()));
    connect(optionsClassify->crossValidButton, SIGNAL(clicked()), this, SLOT(ClassifyCross()));

    connect(optionsRegress->regressionButton, SIGNAL(clicked()), this, SLOT(Regression()));
    connect(optionsRegress->crossValidButton, SIGNAL(clicked()), this, SLOT(RegressionCross()));
    connect(optionsRegress->clearButton, SIGNAL(clicked()), this, SLOT(Clear()));
    //connect(optionsRegress->svmTypeCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(ChangeActiveOptions()));

    connect(optionsCluster->clusterButton, SIGNAL(clicked()), this, SLOT(Cluster()));
    connect(optionsCluster->iterationButton, SIGNAL(clicked()), this, SLOT(ClusterIterate()));
    connect(optionsCluster->clearButton, SIGNAL(clicked()), this, SLOT(Clear()));

    connect(optionsDynamic->regressionButton, SIGNAL(clicked()), this, SLOT(Dynamize()));
    connect(optionsDynamic->clearButton, SIGNAL(clicked()), this, SLOT(Clear()));
    connect(optionsDynamic->centerCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(ChangeActiveOptions()));
    connect(optionsDynamic->resampleCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(ChangeActiveOptions()));
    connect(optionsDynamic->resampleSpin, SIGNAL(valueChanged(int)), this, SLOT(ChangeActiveOptions()));
    connect(optionsDynamic->dtSpin, SIGNAL(valueChanged(double)), this, SLOT(ChangeActiveOptions()));

    connect(drawToolbar->singleButton, SIGNAL(clicked()), this, SLOT(DrawSingle()));
    connect(drawToolbar->sprayButton, SIGNAL(clicked()), this, SLOT(DrawSpray()));
    connect(drawToolbar->lineButton, SIGNAL(clicked()), this, SLOT(DrawLine()));
    connect(drawToolbar->ellipseButton, SIGNAL(clicked()), this, SLOT(DrawEllipse()));
    connect(drawToolbar->eraseButton, SIGNAL(clicked()), this, SLOT(DrawErase()));
    connect(drawToolbar->trajectoryButton, SIGNAL(clicked()), this, SLOT(DrawTrajectory()));

    connect(drawToolbar->singleButton, SIGNAL(clicked()), this, SLOT(DrawSingle()));
    connect(drawToolbar->sprayButton, SIGNAL(clicked()), this, SLOT(DrawSpray()));
    connect(drawToolbar->lineButton, SIGNAL(clicked()), this, SLOT(DrawLine()));
    connect(drawToolbar->ellipseButton, SIGNAL(clicked()), this, SLOT(DrawEllipse()));
    connect(drawToolbar->eraseButton, SIGNAL(clicked()), this, SLOT(DrawErase()));
    connect(drawToolbar->trajectoryButton, SIGNAL(clicked()), this, SLOT(DrawTrajectory()));

    connect(displayOptions->screenshotButton, SIGNAL(clicked()), this, SLOT(Screenshot()));
    connect(displayOptions->clipboardButton, SIGNAL(clicked()), this, SLOT(ToClipboard()));
    connect(displayOptions->mapCheck, SIGNAL(clicked()), this, SLOT(DisplayOptionChanged()));
    connect(displayOptions->modelCheck, SIGNAL(clicked()), this, SLOT(DisplayOptionChanged()));
    connect(displayOptions->infoCheck, SIGNAL(clicked()), this, SLOT(DisplayOptionChanged()));
    connect(displayOptions->samplesCheck, SIGNAL(clicked()), this, SLOT(DisplayOptionChanged()));
    //	connect(displayOptions->trajectoriesCheck, SIGNAL(clicked()), this, SLOT(DisplayOptionChanged()));
    //	connect(displayOptions->singleclassCheck, SIGNAL(clicked()), this, SLOT(DisplayOptionChanged()));

    optionsClassify->tabWidget->clear();
    optionsCluster->tabWidget->clear();
    optionsRegress->tabWidget->clear();
    optionsDynamic->tabWidget->clear();
    optionsClassify->tabWidget->setUsesScrollButtons(true);
    optionsCluster->tabWidget->setUsesScrollButtons(true);
    optionsRegress->tabWidget->setUsesScrollButtons(true);
    optionsDynamic->tabWidget->setUsesScrollButtons(true);

    canvas = new Canvas(ui.centralWidget);
    canvas->setMouseTracking(true);
    connect(canvas, SIGNAL(Drawing(fvec,int)), this, SLOT(Drawing(fvec,int)));
    connect(canvas, SIGNAL(DrawCrosshair()), this, SLOT(DrawCrosshair()));
    connect(canvas, SIGNAL(Navigation(fvec)), this, SLOT(Navigation(fvec)));
    connect(canvas, SIGNAL(Released()), this, SLOT(DrawingStopped()));
    drawTimer = new DrawTimer(canvas, &mutex);
    drawTimer->classifier = &classifier;
    drawTimer->regressor = &regressor;
    drawTimer->dynamical = &dynamical;
    drawTimer->clusterer = &clusterer;
    connect(drawTimer, SIGNAL(MapReady(QImage)), canvas, SLOT(SetConfidenceMap(QImage)));
    connect(drawTimer, SIGNAL(ModelReady(QImage)), canvas, SLOT(SetModelImage(QImage)));
}

void MLDemos::initPlugins()
{
    qDebug() << "Importing plugins";
    QDir pluginsDir = QDir(qApp->applicationDirPath());
    QStringList pluginFileNames;
	QDir alternativeDir = pluginsDir;

#if defined(Q_OS_WIN)
    if (pluginsDir.dirName().toLower() == "debug" || pluginsDir.dirName().toLower() == "release") pluginsDir.cdUp();
#elif defined(Q_OS_MAC)
    if (pluginsDir.dirName() == "MacOS") {
        pluginsDir.cdUp();
		if(pluginsDir.cd("Resources/plugins"))
		{
			alternativeDir = pluginsDir;
			pluginsDir.cdUp();
			pluginsDir.cdUp();
		}
        pluginsDir.cdUp();
        pluginsDir.cdUp();
    }
#endif
	bool bFoundPlugins = false;
#if defined(DEBUG)
	bFoundPlugins = pluginsDir.cd("pluginsDebug");
#else
	bFoundPlugins = pluginsDir.cd("plugins");
#endif
	if(!bFoundPlugins)
	{
		qDebug() << "using alternative directory: " << alternativeDir.absolutePath();
		pluginsDir = alternativeDir;
	}
	foreach (QString fileName, pluginsDir.entryList(QDir::Files))
    {
        qDebug() << fileName;
        QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
        QObject *plugin = loader.instance();
        if (plugin)
        {
            // check type of plugin
            CollectionInterface *iCollection = qobject_cast<CollectionInterface *>(plugin);
            if(iCollection)
            {
                std::vector<ClassifierInterface*> classifierList = iCollection->GetClassifiers();
                std::vector<ClustererInterface*> clustererList = iCollection->GetClusterers();
                std::vector<RegressorInterface*> regressorList = iCollection->GetRegressors();
                std::vector<DynamicalInterface*> dynamicalList = iCollection->GetDynamicals();
                FOR(i, classifierList.size()) AddPlugin(classifierList[i], SLOT(ChangeActiveOptions));
                FOR(i, clustererList.size()) AddPlugin(clustererList[i], SLOT(ChangeActiveOptions));
                FOR(i, regressorList.size()) AddPlugin(regressorList[i], SLOT(ChangeActiveOptions));
                FOR(i, dynamicalList.size()) AddPlugin(dynamicalList[i], SLOT(ChangeActiveOptions));
            }
            ClassifierInterface *iClassifier = qobject_cast<ClassifierInterface *>(plugin);
            if (iClassifier)
            {
                AddPlugin(iClassifier, SLOT(ChangeActiveOptions()));
            }
            ClustererInterface *iClusterer = qobject_cast<ClustererInterface *>(plugin);
            if (iClusterer)
            {
                AddPlugin(iClusterer, SLOT(ChangeActiveOptions()));
            }
            RegressorInterface *iRegressor = qobject_cast<RegressorInterface *>(plugin);
            if (iRegressor)
            {
                AddPlugin(iRegressor, SLOT(ChangeActiveOptions()));
            }
            DynamicalInterface *iDynamical = qobject_cast<DynamicalInterface *>(plugin);
            if (iDynamical)
            {
                AddPlugin(iDynamical, SLOT(ChangeActiveOptions()));
            }
            InputOutputInterface *iIO = qobject_cast<InputOutputInterface *>(plugin);
            if (iIO)
            {
                AddPlugin(iIO);
            }
        }
    }
}

void MLDemos::ShowContextMenuSpray(const QPoint &point)
{
    QPoint pt = QPoint(30, 0);
    drawContext1Widget->move(drawToolbar->sprayButton->mapToGlobal(pt));
    drawContext1Widget->show();
    drawContext1Widget->setFocus();
}
void MLDemos::ShowContextMenuLine(const QPoint &point)
{
    QPoint pt = QPoint(30, 0);
    drawContext2Widget->move(drawToolbar->lineButton->mapToGlobal(pt));
    drawContext2Widget->show();
    drawContext2Widget->setFocus();
}
void MLDemos::ShowContextMenuEllipse(const QPoint &point)
{
    QPoint pt = QPoint(30, 0);
    drawContext2Widget->move(drawToolbar->ellipseButton->mapToGlobal(pt));
    drawContext2Widget->show();
    drawContext2Widget->setFocus();
}
void MLDemos::ShowContextMenuErase(const QPoint &point)
{
    QPoint pt = QPoint(30, 0);
    drawContext1Widget->move(drawToolbar->eraseButton->mapToGlobal(pt));
    drawContext1Widget->show();
    drawContext1Widget->setFocus();
}

bool IsChildOf(QObject *child, QObject *parent)
{
    if(!parent || !child) return false;
    qDebug() << "testing: " << parent << " and " << child;
    if(child == parent) return true;
    QList<QObject*> list = parent->children();
    if(list.isEmpty()) return false;
    QList<QObject*>::iterator i;
    for (i = list.begin(); i<list.end(); ++i)
    {
        if(IsChildOf(child, *i)) return true;
    }
    return false;
}

void MLDemos::FocusChanged(QWidget *old, QWidget *now)
{
    if(drawContext1Widget->isVisible())
    {
        if(!IsChildOf(now, drawContext1Widget)) HideContextMenus();
    }
    if(drawContext2Widget->isVisible())
    {
        if(!IsChildOf(now, drawContext2Widget)) HideContextMenus();
    }
}

void MLDemos::HideContextMenus()
{
    drawContext1Widget->hide();
    drawContext2Widget->hide();
}

void MLDemos::AddPlugin(InputOutputInterface *iIO)
{
    inputoutputs.push_back(iIO);
    connect(this, SIGNAL(SendResults(std::vector<fvec>)), iIO->object(), iIO->FetchResultsSlot());
    connect(iIO->object(), iIO->SetDataSignal(), this, SLOT(SetData(std::vector<fvec>, ivec, std::vector<ipair>)));
    connect(iIO->object(), iIO->QueryClassifierSignal(), this, SLOT(QueryClassifier(std::vector<fvec>)));
    connect(iIO->object(), iIO->QueryRegressorSignal(), this, SLOT(QueryRegressor(std::vector<fvec>)));
    connect(iIO->object(), iIO->QueryDynamicalSignal(), this, SLOT(QueryDynamical(std::vector<fvec>)));
    connect(iIO->object(), iIO->QueryClustererSignal(), this, SLOT(QueryClusterer(std::vector<fvec>)));
    connect(iIO->object(), iIO->DoneSignal(), this, SLOT(DisactivateIO(QObject *)));
    QString name = iIO->GetName();
    QAction *pluginAction = ui.menuInput_Output->addAction(name);
    pluginAction->setCheckable(true);
    pluginAction->setChecked(false);
    connect(pluginAction,SIGNAL(toggled(bool)), this, SLOT(ActivateIO()));
}

void MLDemos::AddPlugin(ClassifierInterface *iClassifier, const char *method)
{
    if(!iClassifier) return;
    // we add the interface so we can use it to produce classifiers
    classifiers.push_back(iClassifier);
    // we add the classifier parameters to the gui
    optionsClassify->tabWidget->addTab(iClassifier->GetParameterWidget(), iClassifier->GetName());
}

void MLDemos::AddPlugin(ClustererInterface *iCluster, const char *method)
{
    if(!iCluster) return;
    clusterers.push_back(iCluster);
    optionsCluster->tabWidget->addTab(iCluster->GetParameterWidget(), iCluster->GetName());
}

void MLDemos::AddPlugin(RegressorInterface *iRegress, const char *method)
{
    if(!iRegress) return;
    regressors.push_back(iRegress);
    optionsRegress->tabWidget->addTab(iRegress->GetParameterWidget(), iRegress->GetName());
}

void MLDemos::AddPlugin(DynamicalInterface *iDynamical, const char *method)
{
    if(!iDynamical) return;
    dynamicals.push_back(iDynamical);
    optionsDynamic->tabWidget->addTab(iDynamical->GetParameterWidget(), iDynamical->GetName());
}

MLDemos::~MLDemos()
{
    SaveLayoutOptions();
    delete optionsClassify;
    delete optionsRegress;
    delete optionsCluster;
    delete optionsDynamic;
    delete drawToolbar;
    delete drawToolbarContext1;
    delete drawToolbarContext2;
    delete displayOptions;
    /*
 FOR(i, classifiers.size()) DEL(classifiers[i]);
 FOR(i, regressors.size()) DEL(regressors[i]);
 FOR(i, dynamicals.size()) DEL(dynamicals[i]);
 FOR(i, clusterers.size()) DEL(clusterers[i]);
 FOR(i, inputoutputs.size()) DEL(inputoutputs[i]);
 */
    canvas->hide();
    delete canvas;
}

void MLDemos::closeEvent(QCloseEvent *event)
{
    if (true)
    {
        mutex.lock();
        DEL(regressor);
        DEL(classifier);
        mutex.unlock();
        qApp->quit();
    } else {
        event->ignore();
    }
}

void MLDemos::resizeEvent( QResizeEvent *event )
{
#ifdef MACX // ugly hack to avoid resizing problems on the mac
    //	qDebug() << "resizing: " << ui.centralWidget->geometry() << endl;
    if(height() < 600) resize(width(),600);
#endif // MACX
    if(canvas)
    {
        //		qDebug() << canvas->geometry() << endl;
        //        qDebug() << ui.centralWidget->geometry();
        int w = ui.centralWidget->width();
        int h = ui.centralWidget->height()-2;
        canvas->setGeometry(0, 0, w, h);
        canvas->ResetSamples();
        char text[255];
        sprintf(text, "resizing: %d x %d", w, h);
        ui.statusBar->showMessage(text);
    }
}

void MLDemos::ShowOptionClass()
{
    if(actionClassifiers->isChecked()) classifyDialog->show();
    else classifyDialog->hide();
    regressDialog->hide(); actionRegression->setChecked(false);
    dynamicDialog->hide(); actionDynamical->setChecked(false);
    clusterDialog->hide(); actionClustering->setChecked(false);
}

void MLDemos::ShowOptionRegress()
{
    if(actionRegression->isChecked()) regressDialog->show();
    else regressDialog->hide();
    classifyDialog->hide(); actionClassifiers->setChecked(false);
    dynamicDialog->hide(); actionDynamical->setChecked(false);
    clusterDialog->hide(); actionClustering->setChecked(false);
}

void MLDemos::ShowOptionDynamical()
{
    if(actionDynamical->isChecked()) dynamicDialog->show();
    else dynamicDialog->hide();
    classifyDialog->hide(); actionClassifiers->setChecked(false);
    regressDialog->hide(); actionRegression->setChecked(false);
    clusterDialog->hide(); actionClustering->setChecked(false);
}

void MLDemos::ShowOptionCluster()
{
    if(actionClustering->isChecked()) clusterDialog->show();
    else clusterDialog->hide();
    classifyDialog->hide(); actionClassifiers->setChecked(false);
    regressDialog->hide(); actionRegression->setChecked(false);
    dynamicDialog->hide(); actionDynamical->setChecked(false);
}

void MLDemos::ShowSampleDrawing()
{
    if(actionDrawSamples->isChecked())
    {
        drawToolbarWidget->show();
    }
    else
    {
        drawToolbarWidget->hide();
    }
}

void MLDemos::ShowOptionDisplay()
{
    if(actionDisplayOptions->isChecked()) displayDialog->show();
    else displayDialog->hide();
}

void MLDemos::ShowStatsDialog()
{
    if(actionShowStats->isChecked()) statsDialog->show();
    else statsDialog->hide();
}

void MLDemos::ShowAbout()
{
    about->show();
}

void MLDemos::HideOptionClass()
{
    classifyDialog->hide();
    actionClassifiers->setChecked(false);
}

void MLDemos::HideOptionRegress()
{
    regressDialog->hide();
    actionRegression->setChecked(false);
}

void MLDemos::HideOptionDynamical()
{
    dynamicDialog->hide();
    actionDynamical->setChecked(false);
}

void MLDemos::HideOptionCluster()
{
    clusterDialog->hide();
    actionClustering->setChecked(false);
}

void MLDemos::HideSampleDrawing()
{
    drawToolbarWidget->hide();
    actionDrawSamples->setChecked(false);
}

void MLDemos::HideOptionDisplay()
{
    displayDialog->hide();
    actionDisplayOptions->setChecked(false);
}

void MLDemos::HideStatsDialog()
{
    statsDialog->hide();
    actionShowStats->setChecked(false);
}


void MLDemos::Clear()
{
    QMutexLocker lock(&mutex);
    DEL(classifier);
    DEL(regressor);
    DEL(dynamical);
    DEL(clusterer);
    drawTimer->Clear();
    canvas->confidencePixmap = QPixmap();
    canvas->modelPixmap = QPixmap();
    canvas->infoPixmap = QPixmap();
    canvas->liveTrajectory.clear();
    canvas->repaint();
    UpdateInfo();
}

void MLDemos::ResetPositiveClass()
{
    int labMin = INT_MAX, labMax = INT_MIN;
    if(!canvas->data->GetCount())
    {
        labMin = 0;
        labMax = 1;
    }
    else
    {
        ivec labels = canvas->data->GetLabels();
        FOR(i, labels.size())
        {
            if(labels[i] > labMax) labMax = labels[i];
            if(labels[i] < labMin) labMin = labels[i];
        }
    }
    optionsClassify->positiveSpin->setRange(labMin,labMax);
    if(optionsClassify->positiveSpin->value() < labMin)
        optionsClassify->positiveSpin->setValue(labMin);
    else if(optionsClassify->positiveSpin->value() > labMax)
        optionsClassify->positiveSpin->setValue(labMax);
}

void MLDemos::ChangeActiveOptions()
{
    /*
 optionsDynamic->svmCLabel->setText("C");
 optionsDynamic->svmPSpin->setRange(0.0001, 1.0);
 optionsDynamic->svmPSpin->setSingleStep(0.01);
 optionsDynamic->svmPSpin->setDecimals(4);
 optionsDynamic->svmCSpin->setEnabled(true);
 optionsDynamic->svmCSpin->setRange(0.1, 9999.9);
 optionsDynamic->svmCSpin->setDecimals(1);
 switch(optionsDynamic->svmTypeCombo->currentIndex())
 {
 case 0: // C-SVM
  optionsDynamic->svmEpsLabel->setText("eps");
  break;
 case 1: // Nu-SVM
  optionsDynamic->svmEpsLabel->setText("Nu");
  break;
 case 2: // SOGP
  optionsDynamic->svmEpsLabel->setText("Noise");
  optionsDynamic->svmCLabel->setText("Capacity");
  optionsDynamic->svmCSpin->setRange(-1, 500);
  optionsDynamic->svmCSpin->setDecimals(0);
  optionsDynamic->svmPSpin->setRange(0.001, 1.0);
  optionsDynamic->svmPSpin->setSingleStep(0.01);
  optionsDynamic->svmPSpin->setDecimals(3);
  break;
 }
 */

    DisplayOptionChanged();
}

void MLDemos::ClearData()
{
    if(canvas) canvas->data->Clear();
    Clear();
    ResetPositiveClass();
    UpdateInfo();
}

void MLDemos::DrawSingle()
{
    if(drawToolbar->singleButton->isChecked())
    {
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }

    if(drawToolbar->singleButton->isChecked())
    {
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }
}

void MLDemos::DrawSpray()
{
    if(drawToolbar->sprayButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }

    if(drawToolbar->sprayButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }
}

void MLDemos::DrawErase()
{
    if(drawToolbar->eraseButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }

    if(drawToolbar->eraseButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
        if(optionsDynamic) optionsDynamic->centerCombo->setCurrentIndex(0);
    }
}

void MLDemos::DrawLine()
{
    if(drawToolbar->lineButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }

    if(drawToolbar->lineButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }
}

void MLDemos::DrawEllipse()
{
    if(drawToolbar->ellipseButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }

    if(drawToolbar->ellipseButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->trajectoryButton->setChecked(false);
    }
}

void MLDemos::DrawTrajectory()
{
    if(drawToolbar->trajectoryButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
    }

    if(drawToolbar->trajectoryButton->isChecked())
    {
        drawToolbar->singleButton->setChecked(false);
        drawToolbar->sprayButton->setChecked(false);
        drawToolbar->eraseButton->setChecked(false);
        drawToolbar->lineButton->setChecked(false);
        drawToolbar->ellipseButton->setChecked(false);
    }
}

void MLDemos::DisplayOptionChanged()
{
    if(!canvas) return;
    canvas->bDisplayInfo = displayOptions->infoCheck->isChecked();
    canvas->bDisplayLearned = displayOptions->modelCheck->isChecked();
    canvas->bDisplayMap = displayOptions->mapCheck->isChecked();
    canvas->bDisplaySamples = displayOptions->samplesCheck->isChecked();
    canvas->bDisplayTrajectories = displayOptions->samplesCheck->isChecked();
    //	canvas->bDisplayTrajectories = displayOptions->trajectoriesCheck->isChecked();
    if(optionsDynamic)
    {
        canvas->trajectoryCenterType = optionsDynamic->centerCombo->currentIndex();
        canvas->trajectoryResampleType = optionsDynamic->resampleCombo->currentIndex();
        canvas->trajectoryResampleCount = optionsDynamic->resampleSpin->value();
    }
    canvas->ResetSamples();
    canvas->repaint();
}

void MLDemos::DrawCrosshair()
{
    int drawType = 0;
    if(drawToolbar->singleButton->isChecked()) drawType = 1;
    if(drawToolbar->sprayButton->isChecked()) drawType = 2;
    if(drawToolbar->eraseButton->isChecked()) drawType = 3;
    if(drawToolbar->ellipseButton->isChecked()) drawType = 4;
    if(drawToolbar->lineButton->isChecked()) drawType = 5;
    if(drawToolbar->trajectoryButton->isChecked()) drawType = 6;

    if(!drawType || drawType == 1 || drawType == 6)
    {
        canvas->crosshair = QPainterPath();
        canvas->bNewCrosshair = false;
        return;
    }
    int type = drawToolbarContext1->randCombo->currentIndex();
    float aX = drawToolbarContext2->spinSigmaX->value();
    float aY = drawToolbarContext2->spinSigmaY->value();
    float angle = -drawToolbarContext2->spinAngle->value()/180.f*PIf;
    float s = drawToolbarContext1->spinSize->value();
    int size = (int)(s*max(canvas->width(), canvas->height()));
    int sizeX = (int)(aX*max(canvas->width(), canvas->height()));
    int Size = max(canvas->width(), canvas->height());

    QPainterPath cursor;

    float sin_angle = sinf(angle);
    float cos_angle = cosf(angle);

    if(drawType == 5)
    {
        QPointF pStart, pStop;
        float x = cos_angle*aX;
        float y = sin_angle*aX;
        pStart = QPointF(- x*Size, - y*Size);
        pStop = QPointF(+ x*Size, + y*Size);
        cursor.moveTo(pStart);
        cursor.lineTo(pStop);
        canvas->crosshair = cursor;
        canvas->bNewCrosshair = false;
        return;
    }
    if(drawType == 2)
    {
        cursor.addEllipse(QPoint(0,0),size/2,size/2);
        canvas->crosshair = cursor;
        canvas->bNewCrosshair = false;
        return;
    }

    if(drawType == 3)
    {
        cursor.addEllipse(QPoint(0,0),size/2,size/2);
        canvas->crosshair = cursor;
        canvas->bNewCrosshair = false;
        return;
    }

    QPointF oldPoint, point;
    for(float theta=0; theta < 2*PIf + 0.1; theta += 0.1f)
    {
        float X, Y;
        if(drawType == 2 || drawType == 3)
        {
            X = sqrtf(aX)/2 * cosf(theta);
            Y = sqrtf(aY)/2 * sinf(theta);
        }
        else
        {
            X = aX * cosf(theta);
            Y = aY * sinf(theta);
        }

        float RX = + X * cos_angle + Y * sin_angle;
        float RY = - X * sin_angle + Y * cos_angle;

        point = QPointF(RX*Size,RY*Size);
        if(theta==0)
        {
            cursor.moveTo(point);
            continue;
        }
        cursor.lineTo(point);
        oldPoint = point;
    }
    canvas->crosshair = cursor;
    canvas->bNewCrosshair = false;
}

void MLDemos::Drawing( fvec sample, int label)
{
    int drawType = 0; // none
    if(drawToolbar->singleButton->isChecked()) drawType = 1;
    if(drawToolbar->sprayButton->isChecked()) drawType = 2;
    if(drawToolbar->eraseButton->isChecked()) drawType = 3;
    if(drawToolbar->ellipseButton->isChecked()) drawType = 4;
    if(drawToolbar->lineButton->isChecked()) drawType = 5;
    if(drawToolbar->trajectoryButton->isChecked()) drawType = 6;
    if(!drawType) return;

    if(label) label = drawToolbar->classSpin->value();
    float speed = 6;

    if(drawType == 1) // single samples
    {
        // we don't want to draw too often
        if(drawTime.elapsed() < 50/speed) return; // msec elapsed since last drawing
        canvas->data->AddSample(sample, label);
        canvas->repaint();
    }
    else if(drawType == 2) // spray samples
    {
        // we don't want to draw too often
        if(drawTime.elapsed() < 200/speed) return; // msec elapsed since last drawing
        int type = drawToolbarContext1->randCombo->currentIndex();
        float s = drawToolbarContext1->spinSize->value();
        float size = s*max(canvas->width(), canvas->height());
        int count = drawToolbarContext1->spinCount->value();

        if(type == 0) // uniform
        {
            fvec newSample;
            newSample.resize(2,0);
            FOR(i, count)
            {
                newSample[0] = (rand()/(float)RAND_MAX - 0.5f)*s + sample[0];
                newSample[1] = (rand()/(float)RAND_MAX - 0.5f)*s + sample[1];
                canvas->data->AddSample(newSample, label);
            }
        }
        else // normal
        {
            // we generate the new data
            gaussian gauss;
            gaussian_init(&gauss,2);
            gauss.mean[0] = sample[0];
            gauss.mean[1] = sample[1];
            gauss.covar->_[0] = s*s/9.f*0.5f;
            gauss.covar->_[1] = 0;
            gauss.covar->_[2] = s*s/9.f*0.5f;
            smat_cholesky(gauss.covar, gauss.covar_cholesky);
            fvec newSample;
            newSample.resize(2,0);
            FOR(i, count)
            {
                gaussian_draw(&gauss, &newSample[0]);
                canvas->data->AddSample(newSample, label);
            }
        }
        canvas->repaint();
    }
    else if(drawType == 3) // erase!
    {
        float s = drawToolbarContext1->spinSize->value();
        float size = s*max(canvas->width(), canvas->height());
        QPointF center(sample[0]*canvas->width(), sample[1]*canvas->height());
        canvas->DeleteData(center, size/2);
        canvas->ResetSamples();
        canvas->repaint();
    }
    else if(drawType == 4) // ellipse
    {
        if(drawTime.elapsed() < 200/speed) return; // msec elapsed since last drawing
        float aX = drawToolbarContext2->spinSigmaX->value();
        float aY = drawToolbarContext2->spinSigmaY->value();
        float angle = -drawToolbarContext2->spinAngle->value()/180.f*PIf;
        int count = drawToolbarContext1->spinCount->value()+1;
        float aspectRatio = canvas->width() / (float) canvas->height();
        float sin_angle = sinf(angle);
        float cos_angle = cosf(angle);

        QPointF oldPoint, point;
        float startTheta = rand()/(float)RAND_MAX*2*PIf;
        for(float theta=0; theta < 2*PIf; theta += 2.f*PIf/count)
        {
            float X = aX * cosf(theta+startTheta);
            float Y = aY * sinf(theta+startTheta);

            float RX = + X * cos_angle + Y * sin_angle;
            float RY = - X * sin_angle + Y * cos_angle;

            fvec newSample;
            newSample.resize(2,0);
            newSample[0] = sample[0] + RX;
            newSample[1] = sample[1] + RY * aspectRatio;
            if(theta==0)
            {
                oldPoint = point;
                continue;
            }
            canvas->data->AddSample(newSample, label);

            oldPoint = point;
        }
        canvas->repaint();
    }
    else if(drawType == 5) // line!
    {
        if(drawTime.elapsed() < 200/speed) return; // msec elapsed since last drawing
        float aX = drawToolbarContext2->spinSigmaX->value();
        float angle = -drawToolbarContext2->spinAngle->value()/180.f*PIf;
        int count = drawToolbarContext1->spinCount->value();
        float aspectRatio = canvas->width() / (float) canvas->height();
        float sin_angle = sinf(angle);
        float cos_angle = cosf(angle);

        QPointF pStart, pStop;
        float x = cos_angle*aX;
        float y = sin_angle*aX;
        pStart = QPointF(sample[0] - x, sample[1] - y*aspectRatio);
        pStop = QPointF(sample[0] + x, sample[1] + y*aspectRatio);
        QPointF oldPoint = pStart;
        float start = (rand() / (float)RAND_MAX - 0.5) * (1/(float)count);
        FOR(i,count)
        {
            QPointF point = (pStop - pStart)*((i+1)/(float)count + start) + pStart;
            fvec newSample;
            newSample.resize(2);
            newSample[0] = point.x();
            newSample[1] = point.y();
            canvas->data->AddSample(newSample, label);
            oldPoint = point;
        }
        canvas->repaint();
    }
    else if(drawType == 6) // trajectory!
    {
        if(trajectory.first == -1) // we're starting a trajectory
        {
            trajectory.first = canvas->data->GetCount();
        }
        // we don't want to draw too often
        //if(drawTime.elapsed() < 50/speed) return; // msec elapsed since last drawing
        canvas->data->AddSample(sample, label, _TRAJ);
        trajectory.second = canvas->data->GetCount()-1;
        canvas->repaint();
    }
    drawTime.restart();
    ResetPositiveClass();
    UpdateInfo();
}

void MLDemos::DrawingStopped()
{
    if(trajectory.first != -1)
    {
        // the last point is a duplicate, we take it out
        canvas->data->AddSequence(trajectory);
        canvas->drawnTrajectories = 0;
        trajectory.first = -1;
        canvas->repaint();
    }
}

void MLDemos::Navigation( fvec sample )
{
    QString information;
    char string[255];
    int count = canvas->data->GetCount();
    int pcount = 0, ncount = 0;
    ivec labels = canvas->data->GetLabels();
    FOR(i, labels.size())
    {
        if(labels[i] == 0) ++pcount;
        else ++ncount;
    }
    sprintf(string, "samples: %d (o:%.3d|x:%.3d)", count, pcount, ncount);
    information += QString(string);
    sprintf(string, " | x: %.3f y: %.3f", sample[0], sample[1]);
    information += QString(string);
    QMutexLocker lock(&mutex);
    if(classifier)
    {
        float score = classifier->Test(sample);
        drawTimer->bPaused = false;
        sprintf(string, " | value: %.4f", score);
        information += QString(string);
    }
    else if(dynamical)
    {
        int count = 500;
        std::vector<fvec> trajectory = dynamical->Test(sample, count);
        canvas->liveTrajectory = trajectory;
        canvas->repaint();
    }
    ui.statusBar->showMessage(information);
}

void MLDemos::SaveData()
{
    if(!canvas) return;
    QString filename = QFileDialog::getSaveFileName(this, tr("Save Data"), "", tr("ML Files (*.ml)"));
    if(filename.isEmpty()) return;
    if(!filename.endsWith(".ml")) filename += ".ml";
    Save(filename);
}
void MLDemos :: Save(QString filename)
{
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly))
    {
        ui.statusBar->showMessage("WARNING: Unable to save file");
        return;
    }
    file.close();
    canvas->data->Save(filename.toAscii());
    SaveParams(filename);
    ui.statusBar->showMessage("Data saved successfully");
}

void MLDemos::LoadData()
{
    if(!canvas) return;
    QString filename = QFileDialog::getOpenFileName(this, tr("Load Data"), "", tr("ML Files (*.ml)"));
    if(filename.isEmpty()) return;
    if(!filename.endsWith(".ml")) filename += ".ml";
    Load(filename);
}

void MLDemos::Load(QString filename)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly))
    {
        ui.statusBar->showMessage("WARNING: Unable to open file");
        return;
    }
    file.close();
    ClearData();
    canvas->data->Load(filename.toAscii());
    LoadParams(filename);
    ui.statusBar->showMessage("Data loaded successfully");
    ResetPositiveClass();
    UpdateInfo();
    canvas->repaint();
}

void MLDemos::dragEnterEvent(QDragEnterEvent *event)
{
    QList<QUrl> dragUrl;
    if(event->mimeData()->hasUrls())
    {
        QList<QUrl> urls = event->mimeData()->urls();
        QStringList dataType;
        dataType << ".ml";
        for(int i=0; i<urls.size(); i++)
        {
            QString filename = urls[i].path();
            for(int j=0; j < dataType.size(); j++)
            {
                if(filename.toLower().endsWith(dataType[j]))
                {
                    dragUrl.push_back(urls[i]);
                    break;
                }
            }
        }
        if(dragUrl.size())
        {
            event->acceptProposedAction();
        }
    }
}

void MLDemos::dropEvent(QDropEvent *event)
{
    if(!event->mimeData()->hasUrls()) return;
    FOR(i, event->mimeData()->urls().length())
    {
        QString filename = event->mimeData()->urls()[i].toLocalFile();
        if(filename.toLower().endsWith(".ml"))
        {
            ClearData();
            canvas->data->Load(filename.toAscii());
            LoadParams(filename);
            ui.statusBar->showMessage("Data loaded successfully");
            ResetPositiveClass();
            UpdateInfo();
            canvas->repaint();
        }
    }
    event->acceptProposedAction();
}

void MLDemos::Screenshot()
{
    QString filename = QFileDialog::getSaveFileName(this, tr("Save Screenshot"), "", tr("Images (*.png *.jpg)"));
    if(filename.isEmpty()) return;
    if(!filename.endsWith(".jpg") && !filename.endsWith(".png")) filename += ".png";
    if(!canvas->SaveScreenshot(filename)) ui.statusBar->showMessage("WARNING: Unable to save image");
    else ui.statusBar->showMessage("Image saved successfully");
}

void MLDemos::ToClipboard()
{
    QPixmap screenshot = canvas->GetScreenshot();
    if(screenshot.isNull())
    {
        ui.statusBar->showMessage("WARNING: Nothing to copy to clipboard");
        return;
    }

    QClipboard *clipboard = QApplication::clipboard();
    clipboard->setImage(screenshot.toImage());
    clipboard->setPixmap(screenshot);
    ui.statusBar->showMessage("Image copied successfully to clipboard");

}

/************************************/
/*                                  */
/*      Input Output functions      */
/*                                  */
/************************************/

void MLDemos::ActivateIO()
{
    QList<QAction *> pluginActions = ui.menuInput_Output->actions();
    FOR(i, inputoutputs.size())
    {
        if(i<pluginActions.size() && inputoutputs[i] && pluginActions[i])
        {
            if(pluginActions[i]->isChecked()) inputoutputs[i]->Start();
            else inputoutputs[i]->Stop();
        }
    }
}

void MLDemos::DisactivateIO(QObject *io)
{
    if(!io) return;
    // first we find the right plugin
    int pluginIndex = -1;
    FOR(i, inputoutputs.size())
    {
        if(inputoutputs[i]->object() == io)
        {
            pluginIndex = i;
            break;
        }
    }
    if(pluginIndex == -1)
    {
        statusBar()->showMessage("Unable to unload plugin: ");
        return; // something weird is going on!
    }
    QList<QAction *> pluginActions = ui.menuInput_Output->actions();
    if(pluginIndex < pluginActions.size() && pluginActions[pluginIndex])
    {
        pluginActions[pluginIndex]->setChecked(false);
        inputoutputs[pluginIndex]->Stop();
    }
}

void MLDemos::SetData(std::vector<fvec> samples, ivec labels, std::vector<ipair> trajectories)
{
    canvas->data->Clear();
    canvas->data->AddSamples(samples, labels);
    if(trajectories.size())
    {
        canvas->data->AddSequences(trajectories);
    }
    canvas->ResetSamples();
    canvas->repaint();
}

void MLDemos::QueryClassifier(std::vector<fvec> samples)
{
    std::vector<fvec> results;
    QMutexLocker lock(&mutex);
    fvec result;
    result.resize(1);
    if(classifier && samples.size())
    {
        results.resize(samples.size());
        FOR(i, samples.size())
        {
            result[0] = classifier->Test(samples[i]);
            results[i] = result;
        }
    }
    emit SendResults(results);
}

void MLDemos::QueryRegressor(std::vector<fvec> samples)
{
    std::vector<fvec> results;
    QMutexLocker lock(&mutex);
    if(regressor && samples.size())
    {
        results.resize(samples.size());
        FOR(i, samples.size())
        {
            results[i] = regressor->Test(samples[i]);
        }
    }
    emit SendResults(results);
}

void MLDemos::QueryDynamical(std::vector<fvec> samples)
{
    std::vector<fvec> results;
    QMutexLocker lock(&mutex);
    if(dynamical && samples.size())
    {
        results.resize(samples.size());
        FOR(i, samples.size())
        {
            results[i] = dynamical->Test(samples[i]);
        }
    }
    emit SendResults(results);
}

void MLDemos::QueryClusterer(std::vector<fvec> samples)
{
    std::vector<fvec> results;
    QMutexLocker lock(&mutex);
    if(clusterer && samples.size())
    {
        results.resize(samples.size());
        FOR(i, samples.size())
        {
            results[i] = clusterer->Test(samples[i]);
        }
    }
    emit SendResults(results);
}

